#! /usr/bin/python

""" Main program to synchronize btrfs snapshots.  See README.md.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

if True:  # Headers
    if True:  # imports

        import argparse
        import errno
        import logging
        import logging.handlers
        import os.path
        import re
        import sys

        from util import humanize
        import BestDiffs
        import ButterStore
        import S3Store
        import SSHStore

theDebug = False

logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')

try:
    import version
    theVersion = version.version
except IOError:
    print("Can't import version.py")
    theVersion = "<unknown>"

theChunkSize = 20

command = argparse.ArgumentParser(
    description="Synchronize two sets of btrfs snapshots.",
    epilog="""
<src>, <dst>:   [btrfs://]/path/to/directory/[snapshot]
                s3://bucket/prefix/[snapshot]
                ssh://[user@]host/path/to/directory/[snapshot]

If only <dst> is supplied, just list available snapshots.  NOTE: The trailing
"/" *is* significant.

Copyright (c) 2014-2016 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
See README.md and LICENSE.txt for more info.
    """,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)

command.add_argument('source', metavar='<src>', nargs='?',  # nargs='+',
                     help='a source of btrfs snapshots')
command.add_argument('dest', metavar='<dst>',
                     help='the btrfs snapshots to be updated')

command.add_argument('-n', '--dry-run', action="store_true",
                     help="display what would be transferred, but don't do it",
                     )
command.add_argument('-d', '--delete', action="store_true",
                     help='delete any snapshots in <dst> that are not in <src>',
                     )
command.add_argument('-e', '--estimate', action="store_true",
                     help='use estimated size instead of measuring diffs with a local test send',
                     )

command.add_argument('-q', '--quiet', action="store_true",
                     help='only display error messages',
                     )
command.add_argument('-l', '--logfile', type=argparse.FileType('a'),
                     help='log debugging information to file',
                     )
command.add_argument('-V', '--version', action="version", version='%(prog)s ' + theVersion,
                     help='display version',
                     )

command.add_argument('--part-size', action="store", type=int, default=theChunkSize,
                     help='Size of chunks in a multipart upload',
                     )

# Internals for SSH communication between two buttersinks

command.add_argument('--server', action="store_true",
                     help=argparse.SUPPRESS,
                     )

command.add_argument('--mode',
                     help=argparse.SUPPRESS,
                     )


def _setupLogging(quiet, logFile, isServer):
    theDisplayFormat = '%(message)s'
    theDebugDisplayFormat = (
        '%(levelname)7s:'
        '%(filename)s[%(lineno)d]: %(message)s'
    )
    theLogFormat = (
        '%(asctime)-15s: %(levelname)7s:'
        '%(filename)s[%(lineno)d]: %(message)s'
    )
    theProgram = "buttersink[%d]" % (os.getpid())
    theSysLogFormat = (
        theProgram + ': %(filename)s[%(lineno)d]: %(message)s'
    )

    root = logging.getLogger()
    root.setLevel("INFO" if theDebug else "DEBUG")  # When debugging, this is handled per-logger

    def add(handler, level, format):
        handler = logging.StreamHandler(handler)
        handler.setLevel(level)
        handler.setFormatter(logging.Formatter(format))
        root.addHandler(handler)

    level = "DEBUG" if theDebug else "INFO" if not quiet else "WARN"
    formatString = theDebugDisplayFormat if theDebug else theDisplayFormat
    formatString = ("S|" if isServer else "  ") + formatString

    add(sys.stderr, level, formatString)

    if logFile is not None:
        add(logFile, "DEBUG", theLogFormat)

    if isServer:
        handler = logging.handlers.SysLogHandler(address='/dev/log')
        handler.setLevel(level)
        handler.setFormatter(logging.Formatter(theSysLogFormat))
        root.addHandler(handler)

    logging.getLogger('boto').setLevel("WARN")


def parseSink(uri, isDest, willDelete, dryrun):
    """ Parse command-line description of sink into a sink object. """
    if uri is None:
        return None

    # logger.debug(uri)
    pattern = re.compile('^((?P<method>[^:/]*)://)?(?P<fullpath>(?P<host>[^/]*)(/(?P<path>.*))?)$')
    match = pattern.match(uri)
    if match is None:
        # logger.error("Can't parse snapshot store '%s'", uri)
        raise Exception("Can't parse snapshot store '%s'" % (uri))
    parts = match.groupdict()

    if parts['method'] is None:
        parts['method'] = 'btrfs'

    logger.debug(parts)

    if parts['method'] in ('btrfs', 'file'):
        path = parts['fullpath']
        host = None
    else:
        path = parts['path']
        host = parts['host']

    # Paths specify a directory containing subvolumes,
    # unless it's a source path not ending in "/",
    # then it's a single source subvolume.

    if isDest and not path.endswith("/"):
        path += "/"

    if not isDest:
        mode = 'r'
    elif willDelete:
        mode = 'w'
    else:
        mode = 'a'

    Sinks = {
        'btrfs': ButterStore.ButterStore,
        # 'file': FileStore,
        's3': S3Store.S3Store,
        'ssh': SSHStore.SSHStore,
    }

    return Sinks[parts['method']](host, path, mode, dryrun)


def main():
    """ Main program. """
    try:
        args = command.parse_args()

        _setupLogging(args.quiet, args.logfile, args.server)

        logger.debug("Version: %s, Arguments: %s", theVersion, vars(args))

        if args.server:
            server = SSHStore.StoreProxyServer(args.dest, args.mode)
            return(server.run())

        source = parseSink(args.source, False, args.delete, args.dry_run)

        dest = parseSink(args.dest, source is not None, args.delete, args.dry_run)

        if source is None:
            source = dest
            dest = None

        if not sys.stderr.isatty():
            source.showProgress = dest.showProgress = False
        elif dest is None or (source.isRemote and not dest.isRemote):
            source.showProgress = True
        else:
            dest.showProgress = True

        with source:
            try:
                next(source.listVolumes())
            except StopIteration:
                logger.warn("No snapshots in source.")
                path = args.source or args.dest
                if path.endswith("/"):
                    logger.error(
                        "'%s' does not contain any snapshots.  Did you mean to type '%s'?",
                        path, path[0:-1]
                    )
                else:
                    logger.error(
                        "'%s' is not a snapshot.  Did you mean to type '%s/'?",
                        path, path
                    )
                return 1

            if dest is None:
                for item in source.listContents():
                    print item
                if args.delete:
                    source.deletePartials()
                return 0

            with dest:
                best = BestDiffs.BestDiffs(source.listVolumes(), args.delete, not args.estimate)
                best.analyze(args.part_size << 20, source, dest)

                summary = best.summary()
                logger.info("Optimal synchronization:")
                for sink, values in summary.items():
                    logger.info("%s from %d diffs in %s",
                                humanize(values.size),
                                values.count,
                                sink or "TOTAL",
                                )

                for diff in best.iterDiffs():
                    if diff is None:
                        raise Exception("Missing diff.  Can't fully replicate.")
                    else:
                        diff.sendTo(dest, chunkSize=args.part_size << 20)

                if args.delete:
                    dest.deleteUnused()

        logger.debug("Successful exit")

        return 0
    except Exception as error:
        if (
            isinstance(error, IOError) and
            error.errno == errno.EPERM and
            os.getuid() != 0
        ):
            logger.error("You must be root to access a btrfs filesystem.  Use  'sudo'")
        else:
            if not theDebug:
                logger.debug("Trace information for debugging", exc_info=True)
            logger.error("ERROR: %s.", error, exc_info=theDebug)
        return 1

if __name__ == "__main__":
    sys.exit(main())



""" Manage read-only snapshots in a remote filesystem over SSH.

Copyright (c) 2014-2015 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
"""

from progress import DisplayProgress
import ButterStore
import Store
import version

import io
import json
import logging
import os.path
import platform
import subprocess
import sys
import traceback
import urllib

logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')

DEVNULL = open(os.devnull, 'wb')

theVersion = version.version


class _Obj2Arg:

    def vol(self, vol):
        return 'None' if vol is None else vol.uuid

    def diff(self, diff):
        if diff is None:
            return ('None', 'None')
        else:
            return (self.vol(diff.toVol), self.vol(diff.fromVol), )


class _Arg2Obj:

    def __init__(self, store):
        self.sink = store

    def vol(self, uuid):
        return None if uuid == 'None' else Store.Volume(uuid, None)

    def diff(self, toUUID, fromUUID, estimatedSize='None'):
        if estimatedSize == 'None':
            return Store.Diff(self.sink, self.vol(toUUID), self.vol(fromUUID))
        else:
            estimatedSize = int(float(estimatedSize))
            return Store.Diff(self.sink, self.vol(toUUID), self.vol(fromUUID), estimatedSize, True)

    def bool(self, text):
        return text.lower() in ['true', 'yes', 'on', '1', 't', 'y']


class _Obj2Dict:

    """ Serialize to dictionary. """

    def vol(self, vol):
        if vol is None:
            return None
        return dict(
            uuid=vol.uuid,
            gen=vol.gen,
            size=vol.size,
            exclusiveSize=vol.exclusiveSize,
        )

    def diff(self, diff):
        """ Serialize to a dictionary. """
        if diff is None:
            return None
        return dict(
            toVol=diff.toUUID,
            fromVol=diff.fromUUID,
            size=diff.size,
            sizeIsEstimated=diff.sizeIsEstimated,
        )


class _Dict2Obj:

    def __init__(self, store):
        self.sink = store

    def vol(self, values):
        return Store.Volume(**values)

    def diff(self, values):
        return Store.Diff(sink=self.sink, **values)


class _SSHStream(io.RawIOBase):

    def __init__(self, client, progress=None):
        self._client = client
        self._open = True
        self._progress = progress
        self.totalSize = 0

    def __enter__(self):
        if self._progress:
            self._progress.__enter__()
        return self

    def __exit__(self, exceptionType, exception, trace):
        if self._progress:
            self._progress.__exit__(exceptionType, exception, trace)

        if not self._open:
            return False

        try:
            result = self._client.streamWrite(0)
            self._open = False
        except Exception as error:
            if exceptionType is None:
                raise
            else:
                logger.debug("Secondary error: %s", error)

        if exceptionType is None and result and 'error' in result:
            raise Exception(result)

        return False  # Don't supress exception

    def write(self, data):
        size = len(data)

        if size == 0:
            return

        try:
            result = self._client.streamWrite(size)

            if self._progress:
                self._progress.update(self.totalSize)

            if result.get('stream', False):
                self._client._process.stdin.write(data)
                self.totalSize += size
                if self._progress:
                    self._progress.update(self.totalSize)

                result = self._client._getResult()
        except Exception as error:
            self._client.error = error  # Don't try writing to this client again
            raise

        if result and 'error' in result:
            raise Exception(result)

    def read(self, size):
        if size == 0:
            return ''

        try:
            result = self._client.streamRead(size)

            size = result['size']
            if size == 0:
                self._open = False
                return ''

            if self._progress:
                self._progress.update(self.totalSize)
            data = self._client._process.stdout.read(size)
            self.totalSize += size
            if self._progress:
                self._progress.update(self.totalSize)

            result = self._client._getResult()
        except Exception as error:
            self._client.error = error  # Don't try reading from this client again
            raise

        if result and 'error' in result:
            raise Exception(result)

        return data


class SSHStore(Store.Store):

    """ A synchronization source or sink to a btrfs over SSH. """

    def __init__(self, host, path, mode, dryrun):
        """ Initialize.

        :arg host:   ssh host.
        """
        # For simplicity, only allow absolute paths
        # Don't lose a trailing slash -- it's significant
        path = "/" + os.path.normpath(path) + ("/" if path.endswith("/") else "")

        super(SSHStore, self).__init__(host, path, mode, dryrun)

        self.host = host
        self._client = _Client(host, 'r' if dryrun else mode, path)
        self.isRemote = True

        self.toArg = _Obj2Arg()
        self.toObj = _Dict2Obj(self)

    def __unicode__(self):
        """ English description of self. """
        return u"ssh://%s%s" % (self.host, self.userPath)

    def _open(self):
        """ Open connection to remote host. """
        self._client._open()

    def _close(self):
        """ Close connection to remote host. """
        self._client._close()

    # Abstract methods

    def _fillVolumesAndPaths(self, paths):
        """ Fill in paths.

        :arg paths: = { Store.Volume: ["linux path",]}
        """
        for (volDict, volPaths) in self._client.fillVolumesAndPaths():
            vol = Store.Volume(**volDict)
            paths[vol] = volPaths

    # Abstract methods

    def getEdges(self, fromVol):
        """ Return the edges available from fromVol. """
        return [
            self.toObj.diff(diff)
            for diff in self._client.getEdges(self.toArg.vol(fromVol))
        ]

    def measureSize(self, diff, chunkSize):
        """ Spend some time to get an accurate size. """
        (toUUID, fromUUID) = self.toArg.diff(diff)
        isInteractive = sys.stderr.isatty()
        return self.toObj.diff(self._client.measureSize(
            toUUID,
            fromUUID,
            diff.size,
            chunkSize,
            isInteractive,
        ))

    def hasEdge(self, diff):
        """ True if Store already contains this edge. """
        return diff.toVol in self.paths

    def send(self, diff):
        """ Return Context Manager for a file-like (stream) object to send a diff. """
        if Store.skipDryRun(logger, self.dryrun)("send %s", diff):
            return None

        (diffTo, diffFrom) = self.toArg.diff(diff)
        self._client.send(diffTo, diffFrom)

        progress = DisplayProgress(diff.size) if self.showProgress is True else None
        return _SSHStream(self._client, progress)

    def receive(self, diff, paths):
        """ Return Context Manager for a file-like (stream) object to store a diff. """
        path = self.selectReceivePath(paths)
        path = self._relativePath(path)

        if Store.skipDryRun(logger, self.dryrun)("receive to %s", path):
            return None

        (diffTo, diffFrom) = self.toArg.diff(diff)
        self._client.receive(path, diffTo, diffFrom)

        progress = DisplayProgress(diff.size) if self.showProgress is True else None
        return _SSHStream(self._client, progress)

    def receiveVolumeInfo(self, paths):
        """ Return Context Manager for a file-like (stream) object to store volume info. """
        path = self.selectReceivePath(paths)
        path = path + Store.theInfoExtension

        if Store.skipDryRun(logger, self.dryrun)("receive info to %s", path):
            return None

        self._client.receiveInfo(path)

        return _SSHStream(self._client)

    def keep(self, diff):
        """ Mark this diff (or volume) to be kept in path. """
        (toUUID, fromUUID) = self.toArg.diff(diff)
        self._client.keep(toUUID, fromUUID)
        logger.debug("Kept %s", diff)

    def deleteUnused(self):
        """ Delete any old snapshots in path, if not kept. """
        if self.dryrun:
            self._client.listUnused()
        else:
            self._client.deleteUnused()

    def deletePartials(self):
        """ Delete any old partial uploads/downloads in path. """
        if self.dryrun:
            self._client.listPartials()
        else:
            self._client.deletePartials()


class _Client(object):

    def __init__(self, host, mode, directory):
        self._host = host
        self._mode = mode
        self._directory = urllib.quote_plus(directory, '/')
        self._process = None
        self.error = None

    def _open(self):
        """ Open connection to remote host. """
        if self._process is not None:
            return

        cmd = [
            'ssh',
            self._host,
            'sudo',
            'buttersink',
            '--server',
            '--mode',
            self._mode,
            self._directory
        ]
        logger.debug("Connecting with: %s", cmd)
        self._process = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stderr=sys.stderr,
            # stdout=sys.stdout,
            stdout=subprocess.PIPE,
        )

        version = self.version()
        logger.info("Remote version: %s", version)

    def _close(self):
        """ Close connection to remote host. """
        if self._process is None:
            return

        self.quit()

        self._process.stdin.close()

        logger.debug("Waiting for ssh process to finish...")
        self._process.wait()  # Wait for ssh session to finish.

        # self._process.terminate()
        # self._process.kill()

        self._process = None

    def _checkMode(self, name, mode):
        modes = ["read-only", "append", "write"]

        def value(mode):
            return [s[0] for s in modes].index(mode)

        allowedMode = value(self._mode)
        requestedMode = value(mode)
        if requestedMode > allowedMode:
            raise Exception(
                "%s connection does not allow %s method '%s'" %
                (modes[allowedMode], modes[requestedMode], name)
            )

    def _getResult(self):
        result = self._process.stdout.readline().rstrip("\n")
        # logger.debug('Result: %s', result)
        try:
            result = json.loads(result)
        except:
            # result += os.read(self._process.stdout.fileno(), 5)
            # result += self._process.stdout.read()
            logger.error(result)
            raise Exception("Fatal remote ssh server error")
        return result

    def _sendCommand(self, *command):
        if self.error is not None:
            # logger.warn("Not sending %s because of %s", command, self.error)
            return dict(error=self.error, message="Can't send command", command=command[0])

        try:
            command = ['None' if c is None else urllib.quote_plus(str(c), '/') for c in command]
        except Exception:
            raise Exception("Can't send '%s' over ssh." % (command,))

        # logger.debug("Command line: %s", command)
        commandLine = " ".join(command) + "\n"

        try:
            self._process.stdin.write(commandLine)

            result = self._getResult()
        except Exception as error:
            self.error = error
            raise

        if result and 'error' in result:
            raise Exception(result)

        return result

    @classmethod
    def _addMethod(cls, method, name, mode):
        def fn(self, *args):
            self._checkMode(name, mode)
            return self._sendCommand(name, *args)
        setattr(cls, method, fn)

commands = {}


def command(name, mode):
    """ Label a method as a command with name. """
    def decorator(fn):
        commands[name] = fn.__name__
        _Client._addMethod(fn.__name__, name, mode)
        return fn
    return decorator


class StoreProxyServer(object):

    """ Runs a ButterStore and responds to queries over ssh.

    Use in a 'with' statement.
    """

    def __init__(self, path, mode):
        """ Initialize. """
        logger.debug("Proxy(%s) %s", mode, path)
        self.path = path
        self.mode = mode
        self.butterStore = None
        self.running = False
        self.toObj = None
        self.toDict = None
        self.stream = None

    def __enter__(self):
        """ Enter 'with' statement. """
        return self

    def __exit__(self, exceptionType, exception, trace):
        """ Exit 'with' statement. """
        try:
            self._close()
        except Exception as error:
            if exceptionType is None:
                raise
            else:
                logger.info("Error on close: %s", error)
        return False

    def _open(self, stream):
        if self.stream is not None:
            logger.warn("%s not closed.", self.stream)
        self._close()
        self.stream = stream
        self.stream.__enter__()

    def _close(self):
        if self.stream is None:
            return
        try:
            self.stream.__exit__(None, None, None)
        finally:
            self.stream = None

    def run(self):
        """ Run the server.  Returns with system error code. """
        normalized = os.path.normpath(self.path) + ("/" if self.path.endswith("/") else "")
        if self.path != normalized:
            sys.stderr.write("Please use full path '%s'" % (normalized,))
            return -1

        self.butterStore = ButterStore.ButterStore(None, self.path, self.mode, dryrun=False)
        # self.butterStore.ignoreExtraVolumes = True

        self.toObj = _Arg2Obj(self.butterStore)
        self.toDict = _Obj2Dict()

        self.running = True

        with self.butterStore:
            with self:
                while self.running:
                    self._processCommand()

        return 0

    def _errorInfo(self, command, error):
        trace = traceback.format_exc()
        trace = trace.splitlines()[-3]

        return dict(
            error=str(error),
            errorType=type(error).__name__,
            command=command,
            server=True,
            traceback=trace,
            )

    def _sendResult(self, result):
        """ Send parseable json result of command. """
        # logger.debug("Result: %s", result)

        try:
            result = json.dumps(result)
        except Exception as error:
            result = json.dumps(self._errorInfo(command, error))

        sys.stdout.write(result)
        sys.stdout.write("\n")
        sys.stdout.flush()

    def _errorMessage(self, message):
        sys.stderr.write(str(message))
        sys.stderr.write("\n")
        sys.stderr.flush()

    def _processCommand(self):
        commandLine = sys.stdin.readline().rstrip('\n').split(" ")
        commandLine = [urllib.unquote_plus(c) for c in commandLine]
        command = commandLine[0]

        # logger.debug("Command: '%s'", commandLine)

        if not command:
            self._errorMessage("No more commands -- terminating.")
            self.running = False
            return

        try:
            if command not in commands:
                raise Exception("Unknown command")

            method = getattr(self, commands[command])
            fn = method.__get__(self, StoreProxyServer)
            result = fn(*commandLine[1:])
            if result is None:
                result = dict(command=command, success=True)
        except Exception as error:
            # logger.exception("Failed %s", command)
            result = self._errorInfo(command, error)

        self._sendResult(result)

    @command('quit', 'r')
    def quit(self):
        """ Quit the server. """
        self.running = False
        return dict(message="Quitting")

    @command('version', 'r')
    def version(self):
        """ Return kernel and btrfs version. """
        return dict(
            buttersink=theVersion,
            btrfs=self.butterStore.butter.btrfsVersion,
            linux=platform.platform(),
        )

    @command('send', 'r')
    def send(self, diffTo, diffFrom):
        """ Do a btrfs send. """
        diff = self.toObj.diff(diffTo, diffFrom)
        self._open(self.butterStore.send(diff))

    @command('receive', 'a')
    def receive(self, path, diffTo, diffFrom):
        """ Receive a btrfs diff. """
        diff = self.toObj.diff(diffTo, diffFrom)
        self._open(self.butterStore.receive(diff, [path, ]))

    @command('write', 'r')
    def streamWrite(self, size):
        """ Send or receive a chunk of data.

        :arg size:  Amount of data.  0 indicates EOT.
        """
        size = int(size)
        if size == 0:
            self._close()
            return

        self._sendResult(dict(message="writing...", stream=True, size=size))
        data = sys.stdin.read(size)
        self.stream.write(data)

    @command('read', 'r')
    def streamRead(self, size):
        """ Send or receive a chunk of data.

        :arg size:  Amount of data requested.
        """
        size = int(size)
        data = self.stream.read(size)
        size = len(data)
        if size == 0:
            self._close()
            return dict(message="Finished", size=0)

        self._sendResult(dict(message="reading...", stream=True, size=size))
        sys.stdout.write(data)

    @command('volumes', 'r')
    def fillVolumesAndPaths(self):
        """ Get all volumes for initialization. """
        return [
            (self.toDict.vol(vol), paths)
            for vol, paths in self.butterStore.paths.items()
        ]

    @command('edges', 'r')
    def getEdges(self, fromVol):
        """ Return the edges available from fromVol. """
        return [self.toDict.diff(d) for d in self.butterStore.getEdges(self.toObj.vol(fromVol))]

    @command('measure', 'r')
    def measureSize(self, diffTo, diffFrom, estimatedSize, chunkSize, isInteractive):
        """ Spend some time to get an accurate size. """
        diff = self.toObj.diff(diffTo, diffFrom, estimatedSize)
        isInteractive = self.toObj.bool(isInteractive)
        self.butterStore.showProgress = None if isInteractive else False
        self.butterStore.measureSize(diff, int(chunkSize))
        return self.toDict.diff(diff)

    @command('keep', 'r')
    def keep(self, diffTo, diffFrom):
        """ Mark this diff (or volume) to be kept in path. """
        diff = self.toObj.diff(diffTo, diffFrom)
        self.butterStore.keep(diff)

    @command('delete', 'w')
    def deleteUnused(self):
        """ Delete any old snapshots in path, if not kept. """
        self.butterStore.deleteUnused()

    @command('clean', 'w')
    def deletePartials(self):
        """ Delete any old partial uploads/downloads in path. """
        self.butterStore.deletePartials()

    @command('listDelete', 'r')
    def listUnused(self):
        """ Delete any old snapshots in path, if not kept. """
        self.butterStore.deleteUnused(dryrun=True)

    @command('listClean', 'r')
    def listPartials(self):
        """ Delete any old partial uploads/downloads in path. """
        self.butterStore.deletePartials(dryrun=True)

    @command('info', 'a')
    def receiveInfo(self, path):
        """ Receive volume info. """
        self.stream = open(path, "w")



""" Interface to btrfs-tools for snapshots.

Copyright (c) 2014-2015 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

if True:  # Headers
    if True:  # imports
        import datetime
        import io
        import os
        import os.path
        import psutil
        import re
        import subprocess
        import sys

        from progress import DisplayProgress
        import btrfs
        import send
        import Store

    if True:  # constants
        import logging
        logger = logging.getLogger(__name__)

        DEVNULL = open(os.devnull, 'wb')


# logger.setLevel('DEBUG')

# Btrfs should copy the sending received UUID to the receving UUID,
# but instead it copies the sending current UUID to the receving UUID.
# This prevents diff parents from being properly identified.
# This can be fixed by changing the UUID and transid's duing send/receive.
FIXUP_AFTER_RECEIVE = False
FIXUP_DURING_SEND = True
FIXUP_DURING_RECEIVE = True


def _makeNice(process):
    try:
        ps = psutil.Process(process.pid)
        ps.ionice(psutil.IOPRIO_CLASS_IDLE)
    except AttributeError:
        logger.debug("ionice is not available")


class Butter:

    """ Interface to local btrfs file system snapshots. """

    def __init__(self, dryrun):
        """ Initialize. """
        self.btrfsVersion = self._getVersion([3, 14])
        self.dryrun = dryrun

    def _getVersion(self, minVersion):
        btrfsVersionString = subprocess.check_output(
            ["btrfs", "--version"], stderr=sys.stderr
        ).decode("utf-8").strip()

        versionPattern = re.compile("[0-9]+(\.[0-9]+)*")
        version = versionPattern.search(btrfsVersionString)

        try:
            version = [int(num) for num in version.group(0).split(".")]
        except AttributeError:
            version = None

        if version < [3, 14]:
            logger.error(
                "%s is not supported.  Please upgrade your btrfs to at least 3.14",
                btrfsVersionString
            )
        else:
            logger.debug("%s", btrfsVersionString)

        return btrfsVersionString

    def receive(self, path, diff, showProgress=True):
        """ Return a context manager for stream that will store a diff. """
        directory = os.path.dirname(path)

        cmd = ["btrfs", "receive", "-e", directory]

        if Store.skipDryRun(logger, self.dryrun)("Command: %s", cmd):
            return None

        if not os.path.exists(directory):
            os.makedirs(directory)

        process = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdout=DEVNULL,
        )
        _makeNice(process)

        return _Writer(process, process.stdin, path, diff, showProgress)

    def send(self, targetPath, parent, diff, showProgress=True, allowDryRun=True):
        """ Return context manager for stream to send a (incremental) snapshot. """
        if parent is not None:
            cmd = ["btrfs", "send", "-p", parent, targetPath]
        else:
            cmd = ["btrfs", "send", targetPath]

        if Store.skipDryRun(logger, self.dryrun and allowDryRun)("Command: %s", cmd):
            return None

        process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=DEVNULL)
        _makeNice(process)

        return _Reader(process, process.stdout, targetPath, diff, showProgress)


class _Writer(io.RawIOBase):

    """ Context Manager to write a snapshot. """

    def __init__(self, process, stream, path, diff, showProgress):
        self.process = process
        self.stream = stream
        self.path = path
        self.diff = diff
        self.bytesWritten = None
        self.progress = DisplayProgress(diff.size) if showProgress else None

    def __enter__(self):
        self.bytesWritten = 0
        if self.progress is not None:
            self.progress.open()
        return self

    def __exit__(self, exceptionType, exception, trace):
        self.stream.close()

        if self.progress is not None:
            self.progress.close()

        if self.process is None:
            return

        logger.debug("Waiting for receive process to finish...")
        self.process.wait()

        if exception is None and self.process.returncode == 0:
            # Fixup with SET_RECEIVED_SUBVOL
            if FIXUP_AFTER_RECEIVE:
                received = btrfs.SnapShot(self.path)
                received.SET_RECEIVED_SUBVOL(
                    uuid=self.diff.toUUID,
                    stransid=self.diff.toGen,
                )
            return

        if self.process.returncode != 0:
            logger.error("btrfs receive errors")
            for line in self.process.stderr:
                sys.stderr.write(line)

        if os.path.exists(self.path):
            # This tries to mark partial (failed) transfers.

            partial = self.path + ".part"

            if os.path.exists(partial):
                partial = self.path + "_" + datetime.datetime.now().isoformat() + ".part"

            os.rename(self.path, partial)
            logger.debug("Renamed %s to %s", self.path, partial)

        if exception is None:
            raise Exception(
                "receive %s returned error %d."
                % (self.path, self.process.returncode, )
            )

    def write(self, data):
        # If it's the first big chunk (header)
        # Tweak the volume information to match what we expect.
        if FIXUP_DURING_RECEIVE and self.bytesWritten == 0:
            data = send.replaceIDs(
                data,
                self.diff.toUUID,
                self.diff.toGen,
                self.diff.fromUUID,
                self.diff.fromGen,
            )
        self.stream.write(data)
        self.bytesWritten += len(data)
        if self.progress is not None:
            self.progress.update(self.bytesWritten)


class _Reader(io.RawIOBase):

    """ Context Manager to read a snapshot. """

    def __init__(self, process, stream, path, diff, showProgress):
        self.process = process
        self.stream = stream
        self.path = path
        self.diff = diff
        self.bytesRead = None
        self.progress = DisplayProgress() if showProgress else None

    def __enter__(self):
        self.bytesRead = 0
        if self.progress is not None:
            self.progress.open()
        return self

    def __exit__(self, exceptionType, exception, trace):
        self.stream.close()

        if self.progress is not None:
            self.progress.close()

        if self.process is None:
            return

        logger.debug("Waiting for send process to finish...")
        self.process.wait()

        if self.process.returncode != 0:
            logger.error("btrfs send errors")
            for line in self.process.stderr:
                sys.stderr.write(line)

        if exception is None and self.process.returncode != 0:
            raise Exception(
                "send returned error %d. %s may be corrupt."
                % (self.process.returncode, self.path)
            )

    def read(self, size):
        # If it's the first big chunk (header)
        # Tweak the volume information to match what we expect.
        data = self.stream.read(size)
        if FIXUP_DURING_SEND and self.bytesRead == 0:
            data = send.replaceIDs(
                data,
                self.diff.toUUID,
                self.diff.toGen,
                self.diff.fromUUID,
                self.diff.fromGen,
            )
        self.bytesRead += len(data)
        if self.progress is not None:
            self.progress.update(self.bytesRead)
        return data

    def seek(self, offset, whence):
        self.stream.seek(offset, offset, whence)
        if whence == io.SEEK_SET:
            self.bytesRead = offset
        elif whence == io.SEEK_CUR:
            pass
        elif whence == io.SEEK_END:
            self.bytesRead = None



""" Display Mbs progress on tty. """

from __future__ import division

import util

import datetime
import sys


class DisplayProgress(object):

    """ Class to display Mbs progress on tty. """

    def __init__(self, total=None, chunkName=None, parent=None):
        """ Initialize. """
        self.startTime = None
        self.offset = None
        self.total = total
        self.name = chunkName
        self.parent = parent
        self.output = sys.stderr

    def __enter__(self):
        """ For with statement. """
        self.open()

    def open(self):
        """ Reset time and counts. """
        self.startTime = datetime.datetime.now()
        self.offset = 0
        return self

    def __exit__(self, exceptionType, exceptionValue, traceback):
        """ For with statement. """
        self.close()
        return False

    def update(self, sent):
        """ Update self and parent with intermediate progress. """
        self.offset = sent

        now = datetime.datetime.now()

        elapsed = (now - self.startTime).total_seconds()
        if elapsed > 0:
            mbps = (sent * 8 / (10 ** 6)) / elapsed
        else:
            mbps = None

        self._display(sent, now, self.name, mbps)

    def _display(self, sent, now, chunk, mbps):
        """ Display intermediate progress. """
        if self.parent is not None:
            self.parent._display(self.parent.offset + sent, now, chunk, mbps)
            return

        elapsed = now - self.startTime

        if sent > 0 and self.total is not None and sent <= self.total:
            eta = (self.total - sent) * elapsed.total_seconds() / sent
            eta = datetime.timedelta(seconds=eta)
        else:
            eta = None

        self.output.write(
            "\r %s: Sent %s%s%s ETA: %s (%s) %s%20s\r" % (
                elapsed,
                util.humanize(sent),
                "" if self.total is None else " of %s" % (util.humanize(self.total),),
                "" if self.total is None else " (%d%%)" % (int(100 * sent / self.total),),
                eta,
                "" if not mbps else "%.3g Mbps " % (mbps,),
                chunk or "",
                " ",
            )
        )

        self.output.flush()

    def close(self):
        """ Stop overwriting display, or update parent. """
        if self.parent:
            self.parent.update(self.parent.offset + self.offset)
            return
        self.output.write("\n")
        self.output.flush()



""" Back-end for AWS S3.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

# Docs: http://boto.readthedocs.org/en/latest/

if True:  # Imports and constants
    if True:  # Imports
        from util import humanize
        import progress
        import Store
        import util

        import boto
        import boto.s3.connection
        import collections
        import io
        import logging
        import os.path
        import re
    if True:  # Constants
        # Maximum xumber of progress reports per chunk
        theProgressCount = 50

        # This does transparent S3 server-side encryption
        isEncrypted = True

        # Minimum chuck size is 5M.
        theInfoBufferSize = 10 * (1 << 20)

        logger = logging.getLogger(__name__)

        theTrashPrefix = "trash/"

# logger.setLevel('DEBUG')


def _displayTraceBack():
    return ""
    util.displayTraceBack()


class S3Store(Store.Store):

    """ An S3 bucket synchronization source or sink. """

    def __init__(self, host, path, mode, dryrun):
        """ Initialize.

        host is the bucket name.
        path is an object key prefix to use.

        """
        super(S3Store, self).__init__(host, "/" + path, mode, dryrun)

        self.bucketName = host

        self.keyPattern = re.compile(S3Store.theKeyPattern % ())

        # { fromVol: [diff] }
        self.diffs = None
        self.extraKeys = None

        logger.info("Listing %s contents...", self)

        try:
            # Orginary calling format returns a 301 without specifying a location.
            # Subdomain calling format does not require specifying the region.
            s3 = boto.s3.connection.S3Connection(
                # calling_format=boto.s3.connection.ProtocolIndependentOrdinaryCallingFormat(),
                calling_format=boto.s3.connection.SubdomainCallingFormat(),
                )
            # s3 = boto.connect_s3()   # Often fails with 301
            # s3 = boto.s3.connect_to_region('us-west-2')  # How would we know the region?
        except boto.exception.NoAuthHandlerFound:
            logger.error("Try putting S3 credentials into ~/.boto")
            raise

        self.bucket = s3.get_bucket(self.bucketName)
        self.isRemote = True

    def __unicode__(self):
        """ Return text description. """
        return u'S3 Bucket "%s"' % (self.bucketName)

    def __str__(self):
        """ Return text description. """
        return unicode(self).encode('utf-8')

    def _flushPartialUploads(self, dryrun):
        for upload in self.bucket.list_multipart_uploads():
            # logger.debug("Upload: %s", upload.__dict__)
            # for part in upload:
                # logger.debug("  Part: %s", part.__dict__)

            if not upload.key_name.startswith(self.userPath.lstrip("/")):
                continue

            if self._skipDryRun(logger, 'INFO', dryrun)(
                "Delete old partial upload: %s (%d parts)",
                upload,
                len([part for part in upload]),
            ):
                continue

            upload.cancel_upload()

    def _fillVolumesAndPaths(self, paths):
        """ Fill in paths.

        :arg paths: = { Store.Volume: ["linux path",]}
        """
        self.diffs = collections.defaultdict((lambda: []))
        self.extraKeys = {}

        for key in self.bucket.list():
            if key.name.startswith(theTrashPrefix):
                continue

            keyInfo = self._parseKeyName(key.name)

            if keyInfo is None:
                if key.name[-1:] != '/':
                    logger.warning("Ignoring '%s' in S3", key.name)
                continue

            if keyInfo['type'] == 'info':
                stream = io.BytesIO()
                key.get_contents_to_file(stream)
                Store.Volume.readInfo(stream)
                continue

            if keyInfo['from'] == 'None':
                keyInfo['from'] = None

            path = self._relativePath("/" + keyInfo['fullpath'])

            if path is None:
                continue

            diff = Store.Diff(self, keyInfo['to'], keyInfo['from'], key.size)

            logger.debug("Adding %s in %s", diff, path)

            self.diffs[diff.fromVol].append(diff)
            paths[diff.toVol].append(path)

            self.extraKeys[diff] = path

        # logger.debug("Diffs:\n%s", pprint.pformat(self.diffs))
        # logger.debug("Vols:\n%s", pprint.pformat(self.vols))
        # logger.debug("Extra:\n%s", (self.extraKeys))

    def listContents(self):
        """ Return list of volumes or diffs in this Store's selected directory. """
        items = list(self.extraKeys.items())
        items.sort(key=lambda t: t[1])

        (count, size) = (0, 0)

        for (diff, path) in items:
            if path.startswith("/"):
                continue
            yield str(diff)
            count += 1
            size += diff.size

        yield "TOTAL: %d diffs %s" % (count, humanize(size))

    def getEdges(self, fromVol):
        """ Return the edges available from fromVol. """
        return self.diffs[fromVol]

    def hasEdge(self, diff):
        """ Test whether edge is in this sink. """
        return diff.toVol in [d.toVol for d in self.diffs[diff.fromVol]]

    def measureSize(self, diff, chunkSize):
        """ Spend some time to get an accurate size. """
        logger.warn("Don't need to measure S3 diffs")

    def receive(self, diff, paths):
        """ Return Context Manager for a file-like (stream) object to store a diff. """
        path = self.selectReceivePath(paths)
        keyName = self._keyName(diff.toUUID, diff.fromUUID, path)

        if self._skipDryRun(logger)("receive %s in %s", keyName, self):
            return None

        progress = _BotoProgress(diff.size) if self.showProgress is True else None
        return _Uploader(self.bucket, keyName, progress)

    def receiveVolumeInfo(self, paths):
        """ Return Context Manager for a file-like (stream) object to store volume info. """
        path = self.selectReceivePath(paths)
        path = path + Store.theInfoExtension

        if self._skipDryRun(logger)("receive info in '%s'", path):
            return None

        return _Uploader(self.bucket, path, bufferSize=theInfoBufferSize)

    theKeyPattern = "^(?P<fullpath>.*)/(?P<to>[-a-zA-Z0-9]*)_(?P<from>[-a-zA-Z0-9]*)$"

    def _keyName(self, toUUID, fromUUID, path):
        return "%s/%s_%s" % (self._fullPath(path).lstrip("/"), toUUID, fromUUID)

    def _parseKeyName(self, name):
        """ Returns dict with fullpath, to, from. """
        if name.endswith(Store.theInfoExtension):
            return {'type': 'info'}

        match = self.keyPattern.match(name)
        if not match:
            return None

        match = match.groupdict()
        match.update(type='diff')

        return match

    def send(self, diff):
        """ Write the diff (toVol from fromVol) to the stream context manager. """
        path = self._fullPath(self.extraKeys[diff])
        keyName = self._keyName(diff.toUUID, diff.fromUUID, path)
        key = self.bucket.get_key(keyName)

        if self._skipDryRun(logger)("send %s in %s", keyName, self):
            return None

        progress = _BotoProgress(diff.size) if self.showProgress is True else None
        return _Downloader(key, progress)

    def keep(self, diff):
        """ Mark this diff (or volume) to be kept in path. """
        path = self.extraKeys[diff]

        if not path.startswith("/"):
            logger.debug("Keeping %s", path)
            del self.extraKeys[diff]
            return

        # Copy into self.userPath, if not there already

        keyName = self._keyName(diff.toUUID, diff.fromUUID, path)
        newPath = os.path.join(self.userPath, os.path.basename(path))
        newName = self._keyName(diff.toUUID, diff.fromUUID, newPath)

        if not self._skipDryRun(logger)("Copy %s to %s", keyName, newName):
            self.bucket.copy_key(newName, self.bucket.name, keyName)

    def deleteUnused(self):
        """ Delete any old snapshots in path, if not kept. """
        (count, size) = (0, 0)

        for (diff, path) in self.extraKeys.items():
            if path.startswith("/"):
                continue

            keyName = self._keyName(diff.toUUID, diff.fromUUID, path)

            count += 1
            size += diff.size

            if self._skipDryRun(logger, 'INFO')("Trash: %s", diff):
                continue

            try:
                self.bucket.copy_key(theTrashPrefix + keyName, self.bucket.name, keyName)
                self.bucket.delete_key(keyName)
            except boto.exception.S3ResponseError as error:
                logger.error("%s: %s", error.code, error.message)

            try:
                keyName = os.path.dirname(keyName) + Store.theInfoExtension
                self.bucket.copy_key(theTrashPrefix + keyName, self.bucket.name, keyName)
                self.bucket.delete_key(keyName)
            except boto.exception.S3ResponseError as error:
                logger.debug("%s: %s", error.code, error.message)

        logger.info("Trashed %d diffs (%s)", count, humanize(size))

    def deletePartials(self):
        """ Delete any old partial uploads/downloads in path. """
        self._flushPartialUploads(self.dryrun)


class _BotoProgress(progress.DisplayProgress):

    def __init__(self, total=None, chunkName=None, parent=None):
        super(_BotoProgress, self).__init__(total, chunkName, parent)
        self.numCallBacks = theProgressCount

    def __call__(self, sent, total):
        logger.debug("BotoProgress: %d/%d", sent, total)
        self.total = total
        self.update(sent)

    @staticmethod
    def botoArgs(self):
        if self is not None:
            return {'cb': self, 'num_cb': self.numCallBacks}
        else:
            return {'cb': None, 'num_cb': None}


class _Downloader(io.RawIOBase):

    def __init__(self, key, progress=None):
        self.progress = progress
        self.key = key
        self.mark = 0

    def __enter__(self):
        if self.progress:
            self.progress.__enter__()
        return self

    def __exit__(self, exceptionType, exceptionValue, traceback):
        if self.progress:
            self.progress.__exit__(exceptionType, exceptionValue, traceback)

    def read(self, n=-1):
        if self.mark >= self.key.size or n == 0:
            return b''

        if n < 0:
            headers = None
        else:
            headers = {"Range": "bytes=%s-%s" % (self.mark, self.mark + n - 1)}

        if self.progress is None:
            data = self.key.get_contents_as_string(
                headers,
            )
        else:
            progress = _BotoProgress(n, "Range", self.progress)
            with progress:
                data = self.key.get_contents_as_string(
                    headers, **_BotoProgress.botoArgs(progress)
                )
        size = len(data)

        assert n < 0 or size <= n, (size, data[:10])

        self.mark += size

        return data

    def readable(self):
        return True


class _Uploader(io.RawIOBase):

    def __init__(self, bucket, keyName, progress=None, bufferSize=None):
        self.progress = progress
        self.bucket = bucket
        self.keyName = keyName.lstrip("/")
        self.uploader = None
        self.parts = util.DefaultList()
        self.chunkCount = None
        self.metadata = {}
        self.bufferedWriter = None
        self.bufferSize = bufferSize
        self.exception = None

    def __enter__(self):
        self.open()
        if self.progress:
            self.progress.__enter__()
        if self.bufferSize:
            self.bufferedWriter = io.BufferedWriter(self, buffer_size=self.bufferSize)
            return self.bufferedWriter
        else:
            return self

    def open(self):
        if self.uploader is not None:
            logger.warning(
                "Ignoring double open%s",
                _displayTraceBack(),
            )
            return

        logger.debug("Opening")

        self.chunkCount = 0
        self.parts = util.DefaultList()

        for upload in self.bucket.list_multipart_uploads():
            if upload.key_name != self.keyName:
                continue

            logger.debug("Upload: %s", upload.__dict__)
            for part in upload:
                logger.debug("  Part: %s", part.__dict__)
                self.parts[part.part_number - 1] = (part.size, part.etag)

            if not self.parts:
                continue

            self.uploader = upload
            return

        self.uploader = self.bucket.initiate_multipart_upload(
            self.keyName,
            encrypt_key=isEncrypted,
            metadata=self.metadata,
        )

        assert self.uploader is not None

    def __exit__(self, exceptionType, exceptionValue, traceback):
        self.exception = exceptionValue
        if self.bufferedWriter:
            self.bufferedWriter.close()
            if self.uploader is not None:
                logger.warn("BufferedWriter didn't close raw stream.")
        else:
            self.close()

        if self.progress:
            self.progress.__exit__(exceptionType, exceptionValue, traceback)

        return False  # Don't suppress exception

    def skipChunk(self, chunkSize, checkSum, data=None):
        part = self.parts[self.chunkCount]
        if part is None:
            return False

        (size, tag) = part
        tag = tag.strip('"')

        if size != chunkSize:
            logger.warning("Unexpected chunk size %d instead of %d", chunkSize, size)
            # return False
        if tag != checkSum:
            logger.warning("Bad check sum %d instead of %d", checkSum, tag)
            return False

        self.chunkCount += 1

        logger.info(
            "Skipping already uploaded %s chunk #%d",
            humanize(chunkSize),
            self.chunkCount,
        )

        return True

    def write(self, bytes):
        if len(bytes) == 0:
            logger.debug("Ignoring empty upload request.")
            return 0
        return self.upload(bytes)

    def close(self):
        if self.uploader is None:
            logger.debug(
                "Ignoring double close%s",
                _displayTraceBack(),
            )
            return
        logger.debug(
            "Closing%s",
            _displayTraceBack(),
        )

        if self.exception is None:
            self.uploader.complete_upload()
            # You cannot change metadata after uploading
            # if self.metadata:
            #     key = self.bucket.get_key(self.keyName)
            #     key.update_metadata(self.metadata)
        else:
            # self.uploader.cancel_upload()   # Delete (most of) the uploaded chunks.
            parts = [part for part in self.uploader]
            logger.debug("Uploaded parts: %s", parts)
            pass  # Leave unfinished upload to resume later

        self.uploader = None

    def fileno(self):
        raise IOError("S3 uploads don't use file numbers.")

    def writable(self):
        return True

    def upload(self, bytes):
        if self.chunkCount is None:
            raise Exception("Uploading before opening uploader.")

        self.chunkCount += 1
        size = len(bytes)
        fileObject = io.BytesIO(bytes)

        if self.progress is None:
            self.uploader.upload_part_from_file(
                fileObject, self.chunkCount,
            )
        else:
            cb = _BotoProgress(
                size,
                "Chunk #%d" % (self.chunkCount),
                self.progress,
            )

            with cb:
                self.uploader.upload_part_from_file(
                    fileObject, self.chunkCount, **_BotoProgress.botoArgs(cb)
                )

        return size



""" Manage read-only volumes in local btrfs file system.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

from __future__ import division

import btrfs
import Butter
import progress
import Store

import io
import logging
import math
import os
import os.path
import time

logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')
theMinimumChangeRate = .00001


class ButterStore(Store.Store):

    """ A local btrfs synchronization source or sink. """

    def __init__(self, host, path, mode, dryrun):
        """ Initialize.

        host is ignored.
        path is the file system location of the read-only subvolumes.

        """
        # Don't lose a trailing slash -- it's significant
        path = os.path.abspath(path) + ("/" if path.endswith("/") else "")

        super(ButterStore, self).__init__(host, path, mode, dryrun)

        if not os.path.isdir(self.userPath):
            raise Exception("'%s' is not an existing directory" % (self.userPath))

        self.butter = Butter.Butter(dryrun)  # subprocess command-line interface
        self.btrfs = btrfs.FileSystem(self.userPath)     # ioctl interface

        self.butterVolumes = {}   # Dict of {uuid: <btrfs.Volume>}
        self.extraVolumes = {}  # Will hold volumes inside store directory, but no longer in source

    def _btrfsVol2StoreVol(self, bvol):
        if bvol.received_uuid is not None:
            uuid = bvol.received_uuid
            gen = bvol.sent_gen
        else:
            uuid = bvol.uuid
            gen = bvol.current_gen

        if uuid is None:
            return None

        return Store.Volume(uuid, gen, bvol.totalSize, bvol.exclusiveSize)

    def _fillVolumesAndPaths(self, paths):
        """ Fill in paths.

        :arg paths: = { Store.Volume: ["linux path",]}
        """
        with self.btrfs as mount:
            for bv in mount.subvolumes:
                if not bv.readOnly:
                    continue

                vol = self._btrfsVol2StoreVol(bv)
                if vol is None:
                    continue

                path = bv.fullPath

                if path is None:
                    logger.info("Skipping deleted volume %s", bv.uuid)
                    continue

                relPath = None

                for path in bv.linuxPaths:
                    path = self._relativePath(path)

                    if path is None:
                        continue  # path is outside store scope

                    paths[vol].append(path)

                    infoPath = self._fullPath(path + Store.theInfoExtension)
                    if os.path.exists(infoPath):
                        logger.debug("Reading %s", infoPath)
                        with open(infoPath) as info:
                            Store.Volume.readInfo(info)

                    if not path.startswith("/"):
                        relPath = path

                if vol not in paths:
                    continue

                logger.debug("%s", vol.display(sink=self, detail='phrase'))

                if vol.uuid in self.butterVolumes:
                    logger.warn(
                        "Duplicate effective uuid %s in '%s' and '%s'",
                        vol.uuid, path, self.butterVolumes[vol.uuid].fullPath
                    )

                self.butterVolumes[vol.uuid] = bv

                if relPath is not None:
                    # vol is inside Store directory
                    self.extraVolumes[vol] = relPath

    def _fileSystemSync(self):
        with self.btrfs as mount:
            mount.SYNC()
        time.sleep(2)

    def __unicode__(self):
        """ English description of self. """
        return u"btrfs %s" % (self.userPath)

    def __str__(self):
        """ English description of self. """
        return unicode(self).encode('utf-8')

    def getEdges(self, fromVol):
        """ Return the edges available from fromVol. """
        if fromVol is None:
            for toVol in self.paths:
                yield Store.Diff(self, toVol, fromVol, toVol.size)
            return

        if fromVol not in self.paths:
            return

        fromBVol = self.butterVolumes[fromVol.uuid]
        parentUUID = fromBVol.parent_uuid
        butterDir = os.path.dirname(fromBVol.fullPath)

        vols = [vol for vol in self.butterVolumes.values()
                if vol.parent_uuid == parentUUID or
                os.path.dirname(vol.fullPath) == butterDir
                ]

        changeRate = self._calcChangeRate(vols)

        for toBVol in vols:
            if toBVol == fromBVol:
                continue

            # This gives a conservative estimate of the size of the diff

            estimatedSize = self._estimateSize(toBVol, fromBVol, changeRate)

            toVol = self._btrfsVol2StoreVol(toBVol)

            yield Store.Diff(self, toVol, fromVol, estimatedSize, sizeIsEstimated=True)

    def hasEdge(self, diff):
        """ True if Store already contains this edge. """
        return diff.toUUID in self.butterVolumes

    def receive(self, diff, paths):
        """ Return Context Manager for a file-like (stream) object to store a diff. """
        if not self.dryrun:
            self._fileSystemSync()

        path = self.selectReceivePath(paths)

        if os.path.exists(path):
            raise Exception(
                "Path %s exists, can't receive %s" % (path, diff.toUUID)
            )

        return self.butter.receive(path, diff, self.showProgress is True)

    def receiveVolumeInfo(self, paths):
        """ Return Context Manager for a file-like (stream) object to store volume info. """
        path = self.selectReceivePath(paths)
        path = path + Store.theInfoExtension

        if Store.skipDryRun(logger, self.dryrun)("receive info to %s", path):
            return None

        return open(path, "w")

    def _estimateSize(self, toBVol, fromBVol, changeRate):
        fromGen = fromBVol.current_gen
        genDiff = abs(toBVol.current_gen - fromGen)

        estimatedSize = max(0, toBVol.totalSize - fromBVol.totalSize)
        estimatedSize += toBVol.totalSize * (1 - math.exp(-changeRate * genDiff))
        estimatedSize = max(toBVol.exclusiveSize, estimatedSize)

        return estimatedSize

    def measureSize(self, diff, chunkSize):
        """ Spend some time to get an accurate size. """
        self._fileSystemSync()

        sendContext = self.butter.send(
            self.getSendPath(diff.toVol),
            self.getSendPath(diff.fromVol),
            diff,
            showProgress=self.showProgress is not False,
            allowDryRun=False,
        )

        class _Measure(io.RawIOBase):

            def __init__(self, estimatedSize, showProgress):
                self.totalSize = None
                self.progress = progress.DisplayProgress(estimatedSize) if showProgress else None

            def __enter__(self):
                self.totalSize = 0
                if self.progress:
                    self.progress.__enter__()
                return self

            def __exit__(self, exceptionType, exceptionValue, traceback):
                if self.progress:
                    self.progress.__exit__(exceptionType, exceptionValue, traceback)
                return False  # Don't supress exception

            def writable(self):
                return True

            def write(self, bytes):
                self.totalSize += len(bytes)
                if self.progress:
                    self.progress.update(self.totalSize)

        logger.info("Measuring %s", diff)

        measure = _Measure(diff.size, self.showProgress is not False)
        Store.transfer(sendContext, measure, chunkSize)

        diff.setSize(measure.totalSize, False)

        for path in self.getPaths(diff.toVol):
            path = self._fullPath(path) + Store.theInfoExtension

            with open(path, "a") as infoFile:
                diff.toVol.writeInfoLine(infoFile, diff.fromUUID, measure.totalSize)

    def _calcChangeRate(self, bvols):
        total = 0
        diffs = 0
        minGen = bvols[0].current_gen
        maxGen = minGen
        minSize = bvols[0].totalSize
        maxSize = minSize

        for vol in bvols:
            total += vol.totalSize
            diffs += vol.exclusiveSize
            minGen = min(minGen, vol.current_gen)
            maxGen = max(maxGen, vol.current_gen)
            minSize = min(minSize, vol.totalSize)
            maxSize = max(maxSize, vol.totalSize)

        try:
            # exclusiveSize is often useless,
            # because data may be shared with read-write volumes not usable for send operations
            diffs = max(diffs, maxSize - minSize)
            rate = - math.log(1 - diffs / total) * (len(bvols) - 1) / (maxGen - minGen)
            rate /= 10  # Fudge
        except (ZeroDivisionError, ValueError):
            # logger.debug("Using minimum change rate.")
            rate = theMinimumChangeRate

        # logger.debug("Change rate: %f", rate)

        return rate

    def send(self, diff):
        """ Write the diff (toVol from fromVol) to the stream context manager. """
        if not self.dryrun:
            self._fileSystemSync()

        return self.butter.send(
            self.getSendPath(diff.toVol),
            self.getSendPath(diff.fromVol),
            diff,
            self.showProgress is True,
        )

    def keep(self, diff):
        """ Mark this diff (or volume) to be kept in path. """
        self._keepVol(diff.toVol)
        self._keepVol(diff.fromVol)

    def _keepVol(self, vol):
        """ Mark this volume to be kept in path. """
        if vol is None:
            return

        if vol in self.extraVolumes:
            del self.extraVolumes[vol]
            return

        if vol not in self.paths:
            raise Exception("%s not in %s" % (vol, self))

        paths = self.paths[vol]
        newPath = self.selectReceivePath(paths)
        if self._relativePath(newPath) in paths:
            return

        if self._skipDryRun(logger, 'INFO')("Copy %s to %s", vol, newPath):
            return

        self.butterVolumes[vol.uuid].copy(newPath)

    def deleteUnused(self, dryrun=False):
        """ Delete any old snapshots in path, if not kept. """
        for (vol, path) in self.extraVolumes.items():
            if self._skipDryRun(logger, 'INFO', dryrun=dryrun)("Delete subvolume %s", path):
                continue
            self.butterVolumes[vol.uuid].destroy()

    def deletePartials(self, dryrun=False):
        """ Delete any old partial uploads/downloads in path. """
        for (vol, path) in self.extraVolumes.items():
            if not path.endswith(".part"):
                continue
            if self._skipDryRun(logger, 'INFO', dryrun=dryrun)("Delete subvolume %s", path):
                continue
            self.butterVolumes[vol.uuid].destroy()



""" Utilities for btrfs and buttersink. """

from __future__ import division

import math
import pprint
import traceback


def pretty(obj):
    """ Return pretty representation of obj. """
    # if True:
    #     return pprint.pformat(dict(obj.__dict__))
    return pprint.pformat(obj)
    # try:
    #     return pprint.pformat(dict(obj))
    # except TypeError:
    #     try:
    #         return pprint.pformat(obj.__dict__)
    #     except KeyError:
    #         logger.exception("Funny error.")


def humanize(number):
    """ Return a human-readable string for number. """
    # units = ('bytes', 'KB', 'MB', 'GB', 'TB')
    # base = 1000
    units = ('bytes', 'KiB', 'MiB', 'GiB', 'TiB')
    base = 1024
    if number is None:
        return None
    pow = int(math.log(number, base)) if number > 0 else 0
    pow = min(pow, len(units) - 1)
    mantissa = number / (base ** pow)
    return "%.4g %s" % (mantissa, units[pow])


def displayTraceBack():
    """ Display traceback useful for debugging. """
    tb = traceback.format_stack()
    return "\n" + "".join(tb[:-1])


class DefaultList(list):

    """ list that automatically inserts None for missing items. """

    def __setitem__(self, index, value):
        """ Set item. """
        if len(self) > index:
            return list.__setitem__(self, index, value)
        if len(self) < index:
            self.extend([None] * (index - len(self)))
        list.append(self, value)

    def __getitem__(self, index):
        """ Set item. """
        if index >= len(self):
            return None
        return list.__getitem__(self, index)



""" Abstract and component classes for sources and sinks.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

from util import humanize

import abc
import collections
import functools
import hashlib
import io
import logging
import os.path

theInfoExtension = ".bs"

logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')


class Store(object):

    """ Abstract class for storage back-end.

    Diffs should be indexed by "from" volume.
    Paths should be indexed by volume.

    """

    __metaclass__ = abc.ABCMeta

    # Public methods

    def __init__(self, host, userPath, mode, dryrun):
        """ Initialize. """
        self.host = host

        # { vol: [path] }
        # The order of the paths is important to work around btrfs bugs.
        # The first path is usually the root-volume mounted path,
        # which is required by btrfs 3.14.2.
        self.paths = collections.defaultdict((lambda: []))

        if not userPath.endswith("/"):
            self.userVolume = os.path.basename(userPath)
            userPath = os.path.dirname(userPath)
        else:
            self.userVolume = None

        # This will not end with a "/"
        userPath = os.path.normpath(userPath)

        assert userPath.startswith("/"), userPath
        self.userPath = userPath

        logger.debug("%s('%s')", self.__class__.__name__, userPath)

        self.mode = mode
        self.dryrun = dryrun

        self.ignoreExtraVolumes = False
        self.isRemote = False

        # False - Never show progress
        # True - Always show progress
        # None - Show progress for one-sided actions (e.g. measuring)
        self.showProgress = None

    def __enter__(self):
        """ So we can use a 'with' statement. """
        self._open()
        self._fillVolumesAndPaths(self.paths)
        return self

    def __exit__(self, exceptionType, exceptionValue, traceback):
        """ Clean up after 'with' statement. """
        self._close()
        self.paths = None
        return False  # Don't supress exception

    def listContents(self):
        """ Return list of volumes or diffs in this Store's selected directory. """
        vols = list(self.listVolumes())
        vols.sort(key=lambda v: self.getSendPath(v))
        return [vol.display(self, detail="line") for vol in vols]

    def listVolumes(self):
        """ Return list of all volumes in this Store's selected directory. """
        for (vol, paths) in self.paths.items():
            for path in paths:
                if path.startswith('/'):
                    continue
                if path == '.':
                    continue
                if self.userVolume is not None and os.path.basename(path) != self.userVolume:
                    continue
                yield vol
                break

    def getPaths(self, volume):
        """ Return list of all paths to this volume in this Store. """
        return self.paths[volume]

    def getSendPath(self, volume):
        """ Get a path appropriate for sending the volume from this Store.

        The path may be relative or absolute in this Store.

        """
        try:
            return self._fullPath(next(iter(self.getPaths(volume))))
        except StopIteration:
            return None

    def selectReceivePath(self, paths):
        """ From a set of destination paths, select the best one to receive to.

        The paths are relative or absolute, in a source Store.
        The result will be absolute, suitable for this destination Store.

        """
        logger.debug("%s", paths)
        if not paths:
            path = os.path.basename(self.userPath) + '/Anon'
        try:
            path = [p for p in paths if not p.startswith("/")][0]
        except IndexError:
            path = os.path.basename(list(paths)[0])

        return self._fullPath(path)

    # Private and subclass methods

    def _fullPath(self, path):
        if path.startswith("/"):
            return path
        if path == ".":
            return self.userPath
        return os.path.normpath(os.path.join(self.userPath, path))

    def _relativePath(self, fullPath):
        """ Return fullPath relative to Store directory.

        Return fullPath if fullPath is not inside directory.

        Return None if fullPath is outside our scope.
        """
        if fullPath is None:
            return None

        assert fullPath.startswith("/"), fullPath

        path = os.path.relpath(fullPath, self.userPath)

        if not path.startswith("../"):
            return path
        elif self.ignoreExtraVolumes:
            return None
        else:
            return fullPath

    def _skipDryRun(self, logger, level='DEBUG', dryrun=None):
        return skipDryRun(logger, dryrun or self.dryrun, level)

    def __str__(self):
        """ English description of self.

        Subclasses should just define __unicode__.
        """
        return unicode(self).encode('utf-8')

    # Overridables

    def _open(self):
        """ Open. """
        pass

    def _close(self):
        """ Clean up. """
        pass

    # Abstract methods

    @abc.abstractmethod
    def _fillVolumesAndPaths(self, paths):
        """ Fill in paths.

        :arg paths: = { Store.Volume: ["linux path",]}
        """
        raise NotImplementedError

    # Abstract methods

    @abc.abstractmethod
    def getEdges(self, fromVol):
        """ Return the edges available from fromVol. """
        raise NotImplementedError

    @abc.abstractmethod
    def measureSize(self, diff, chunkSize):
        """ Spend some time to get an accurate size. """
        raise NotImplementedError

    @abc.abstractmethod
    def hasEdge(self, diff):
        """ True if Store already contains this edge. """
        raise NotImplementedError

    @abc.abstractmethod
    def receive(self, diff, paths):
        """ Return Context Manager for a file-like (stream) object to store a diff. """
        raise NotImplementedError

    @abc.abstractmethod
    def send(self, diff):
        """ Return Context Manager for a file-like (stream) object to send a diff. """
        raise NotImplementedError

    @abc.abstractmethod
    def receiveVolumeInfo(self, paths):
        """ Return Context Manager for a file-like (stream) object to store volume info. """
        raise NotImplementedError

    @abc.abstractmethod
    def keep(self, diff):
        """ Mark this diff (or volume) to be kept in path. """
        raise NotImplementedError

    @abc.abstractmethod
    def deleteUnused(self):
        """ Delete any old snapshots in path, if not kept. """
        raise NotImplementedError

    @abc.abstractmethod
    def deletePartials(self):
        """ Delete any old partial uploads/downloads in path. """
        raise NotImplementedError


def transfer(sendContext, receiveContext, chunkSize):
    """ Transfer (large) data from sender to receiver. """
    try:
        chunkSize = receiveContext.chunkSize
    except AttributeError:
        pass

    if sendContext is not None and receiveContext is not None:
        with receiveContext as writer:
            # Open reader after writer,
            # so any raised errors will abort write before writer closes.
            with sendContext as reader:
                checkBefore = None
                if hasattr(writer, 'skipChunk'):
                    checkBefore = hasattr(reader, 'checkSum')

                while True:
                    if checkBefore is True:
                        (size, checkSum) = reader.checkSum(chunkSize)

                        if writer.skipChunk(size, checkSum):
                            reader.seek(size, io.SEEK_CUR)
                            continue

                    data = reader.read(chunkSize)
                    if len(data) == 0:
                        break

                    if checkBefore is False:
                        checkSum = hashlib.md5(data).hexdigest()

                        if writer.skipChunk(len(data), checkSum, data):
                            continue

                    writer.write(data)


class Diff:

    """ Represents a btrfs send diff that creates toVol from fromVol. """

    def __init__(self, sink, toVol, fromVol, size=None, sizeIsEstimated=False):
        """ Initialize. """
        self.sink = sink
        self.toVol = Volume.make(toVol)
        self.fromVol = Volume.make(fromVol)
        self.setSize(size, sizeIsEstimated)

    # {toVolume: {fromVolume: size}}
    theKnownSizes = collections.defaultdict(lambda: collections.defaultdict(lambda: None))

    @property
    def toUUID(self):
        """ 'to' volume's UUID. """
        return self.toVol.uuid

    @property
    def fromUUID(self):
        """ 'from' volume's UUID, if any. """
        return self.fromVol.uuid if self.fromVol else None

    @property
    def toGen(self):
        """ 'to' volume's transid. """
        return self.toVol.gen

    @property
    def fromGen(self):
        """ 'from' volume's transid. """
        return self.fromVol.gen if self.fromVol else None

    @property
    def size(self):
        """ Return size. """
        self._updateSize()
        return self._size

    @property
    def sizeIsEstimated(self):
        """ Return whether size is estimated. """
        self._updateSize()
        return self._sizeIsEstimated

    def setSize(self, size, sizeIsEstimated):
        """ Update size. """
        self._size = size
        self._sizeIsEstimated = sizeIsEstimated

        if self.fromVol is not None and size is not None and not sizeIsEstimated:
            Diff.theKnownSizes[self.toUUID][self.fromUUID] = size

    def sendTo(self, dest, chunkSize):
        """ Send this difference to the dest Store. """
        vol = self.toVol
        paths = self.sink.getPaths(vol)

        if self.sink == dest:
            logger.info("Keep: %s", self)
            self.sink.keep(self)
        else:
            # Log, but don't skip yet, so we can log more detailed skipped actions later
            skipDryRun(logger, dest.dryrun, 'INFO')("Xfer: %s", self)

            receiveContext = dest.receive(self, paths)

            sendContext = self.sink.send(self)

            # try:
            #     receiveContext.metadata['btrfsVersion'] = self.btrfsVersion
            # except AttributeError:
            #     pass

            transfer(sendContext, receiveContext, chunkSize)

        if vol.hasInfo():
            infoContext = dest.receiveVolumeInfo(paths)

            if infoContext is None:
                # vol.writeInfo(sys.stdout)
                pass
            else:
                with infoContext as stream:
                    vol.writeInfo(stream)

    def _updateSize(self):
        if self._size and not self._sizeIsEstimated:
            return

        size = Diff.theKnownSizes[self.toUUID][self.fromUUID]

        if size is None:
            return

        self._size = size
        self._sizeIsEstimated = False

    def __str__(self):
        """ human-readable string. """
        return u"%s from %s (%s%s)" % (
            self.toVol.display(self.sink),
            self.fromVol.display(self.sink) if self.fromVol else "None",
            "~" if self.sizeIsEstimated else "",
            humanize(self.size),
            # self.sink,
        )


class Volume:

    """ Represents a snapshot. """

    def __init__(self, uuid, gen, size=None, exclusiveSize=None):
        """ Initialize. """
        assert uuid is not None
        self._uuid = uuid  # Must never change!
        self.size = size
        self.exclusiveSize = exclusiveSize
        self.gen = gen

    def __cmp__(self, vol):
        """ Compare. """
        return cmp(self._uuid, vol._uuid) if vol else 1

    def __hash__(self):
        """ Hash. """
        return hash(self._uuid)

    @property
    def uuid(self):
        """ Read-only uuid. """
        return self._uuid

    def writeInfoLine(self, stream, fromUUID, size):
        """ Write one line of diff information. """
        if size is None or fromUUID is None:
            return
        if not isinstance(size, int):
            logger.warning("Bad size: %s", size)
            return
        stream.write(str("%s\t%s\t%d\n" % (
            self.uuid,
            fromUUID,
            size,
        )))

    def writeInfo(self, stream):
        """ Write information about diffs into a file stream for use later. """
        for (fromUUID, size) in Diff.theKnownSizes[self.uuid].iteritems():
            self.writeInfoLine(stream, fromUUID, size)

    def hasInfo(self):
        """ Will have information to write. """
        count = len([None
                     for (fromUUID, size)
                     in Diff.theKnownSizes[self.uuid].iteritems()
                     if size is not None and fromUUID is not None
                     ])
        return count > 0

    @staticmethod
    def readInfo(stream):
        """ Read previously-written information about diffs. """
        try:
            for line in stream:
                (toUUID, fromUUID, size) = line.split()
                try:
                    size = int(size)
                except:
                    logger.warning("Bad size: %s", size)
                    continue
                logger.debug("diff info: %s %s %d", toUUID, fromUUID, size)
                Diff.theKnownSizes[toUUID][fromUUID] = size
        except Exception as error:
            logger.warn("Can't read .bs info file (%s)", error)

    def __unicode__(self):
        """ Friendly string for volume. """
        return self.display()

    def __str__(self):
        """ Friendly string for volume. """
        return unicode(self).encode('utf-8')

    def __repr__(self):
        """ Python expression to create self. """
        return "%s(%s)" % (
            self.__class__,
            self.__dict__,
        )

    def display(self, sink=None, detail='phrase'):
        """ Friendly string for volume, using sink paths. """
        if not isinstance(detail, int):
            detail = detailNum[detail]

        if detail >= detailNum['line'] and self.size is not None:
            size = " (%s%s)" % (
                humanize(self.size),
                "" if self.exclusiveSize is None else (
                    " %s exclusive" % (humanize(self.exclusiveSize))
                )
            )
        else:
            size = ""

        vol = "%s %s" % (
            _printUUID(self._uuid, detail - 1),
            sink.getSendPath(self) if sink else "",
        )

        return vol + size

    @classmethod
    def make(cls, vol):
        """ Convert uuid to Volume, if necessary. """
        if isinstance(vol, cls):
            return vol
        elif vol is None:
            return None
        else:
            return cls(vol, None)   # FIX: We need a gen!

detailTypes = ('word', 'phrase', 'line', 'paragraph')
detailNum = {t: n for (n, t) in zip(xrange(len(detailTypes)), detailTypes)}


def display(obj, detail='phrase'):
    """ Friendly string for volume, using sink paths. """
    try:
        return obj.display(detail=detail)
    except AttributeError:
        return str(obj)


def _printUUID(uuid, detail='word'):
    """ Return friendly abbreviated string for uuid. """
    if not isinstance(detail, int):
        detail = detailNum[detail]

    if detail > detailNum['word']:
        return uuid

    if uuid is None:
        return None

    return "%s...%s" % (uuid[:4], uuid[-4:])


def skipDryRun(logger, dryRun, level=logging.DEBUG):
    """ Return logging function.

    When logging function called, will return True if action should be skipped.
    Log will indicate if skipped because of dry run.
    """
    # This is an undocumented "feature" of logging module:
    # logging.log() requires a numeric level
    # logging.getLevelName() maps names to numbers
    if not isinstance(level, int):
        level = logging.getLevelName(level)
    return (
        functools.partial(_logDryRun, logger, level) if dryRun
        else functools.partial(logger.log, level)
    )


def _logDryRun(logger, level, format, *args):
    logger.log(level, "WOULD: " + format % args)
    return True



""" Utilities for sending and receiving btrfs snapshots. """

# Docs: See btrfs-progs/send-utils.c, btrfs-progs/send.h

from ioctl import Structure, t
import btrfs
import ioctl

# import binascii  # This provides "zip" crc
# import crc32c  # This provides *slow* btrfs crc32c
import crcmod.predefined  # This provides fast compiled extension
crc32c = crcmod.predefined.mkPredefinedCrcFun("crc-32c")

import logging
import struct

logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')

BTRFS_SEND_STREAM_MAGIC = "btrfs-stream\0"
BTRFS_SEND_STREAM_VERSION = 1

btrfs_stream_header = Structure(
    (t.char, 'magic', len(BTRFS_SEND_STREAM_MAGIC)),
    (t.le32, 'version'),
    packed=True
)

btrfs_cmd_header = Structure(
    # /* len excluding the header */
    (t.le32, 'len'),
    (t.le16, 'cmd'),
    # /* crc including the header with zero crc field */
    (t.le32, 'crc'),
    packed=True
)

btrfs_tlv_header = Structure(
    (t.le16, 'tlv_type'),
    # /* len excluding the header */
    (t.le16, 'tlv_len'),
    packed=True
)

# /* commands */
(
    BTRFS_SEND_C_UNSPEC,

    BTRFS_SEND_C_SUBVOL,
    BTRFS_SEND_C_SNAPSHOT,

    BTRFS_SEND_C_MKFILE,
    BTRFS_SEND_C_MKDIR,
    BTRFS_SEND_C_MKNOD,
    BTRFS_SEND_C_MKFIFO,
    BTRFS_SEND_C_MKSOCK,
    BTRFS_SEND_C_SYMLINK,

    BTRFS_SEND_C_RENAME,
    BTRFS_SEND_C_LINK,
    BTRFS_SEND_C_UNLINK,
    BTRFS_SEND_C_RMDIR,

    BTRFS_SEND_C_SET_XATTR,
    BTRFS_SEND_C_REMOVE_XATTR,

    BTRFS_SEND_C_WRITE,
    BTRFS_SEND_C_CLONE,

    BTRFS_SEND_C_TRUNCATE,
    BTRFS_SEND_C_CHMOD,
    BTRFS_SEND_C_CHOWN,
    BTRFS_SEND_C_UTIMES,

    BTRFS_SEND_C_END,
    BTRFS_SEND_C_UPDATE_EXTENT,
    __BTRFS_SEND_C_MAX,
) = range(24)
BTRFS_SEND_C_MAX = (__BTRFS_SEND_C_MAX - 1)

# /* attributes in send stream */
(
    BTRFS_SEND_A_UNSPEC,

    BTRFS_SEND_A_UUID,
    BTRFS_SEND_A_CTRANSID,

    BTRFS_SEND_A_INO,
    BTRFS_SEND_A_SIZE,
    BTRFS_SEND_A_MODE,
    BTRFS_SEND_A_UID,
    BTRFS_SEND_A_GID,
    BTRFS_SEND_A_RDEV,
    BTRFS_SEND_A_CTIME,
    BTRFS_SEND_A_MTIME,
    BTRFS_SEND_A_ATIME,
    BTRFS_SEND_A_OTIME,

    BTRFS_SEND_A_XATTR_NAME,
    BTRFS_SEND_A_XATTR_DATA,

    BTRFS_SEND_A_PATH,
    BTRFS_SEND_A_PATH_TO,
    BTRFS_SEND_A_PATH_LINK,

    BTRFS_SEND_A_FILE_OFFSET,
    BTRFS_SEND_A_DATA,

    BTRFS_SEND_A_CLONE_UUID,
    BTRFS_SEND_A_CLONE_CTRANSID,
    BTRFS_SEND_A_CLONE_PATH,
    BTRFS_SEND_A_CLONE_OFFSET,
    BTRFS_SEND_A_CLONE_LEN,

    __BTRFS_SEND_A_MAX,
) = range(26)
BTRFS_SEND_A_MAX = (__BTRFS_SEND_A_MAX - 1)


class ParseException(Exception):

    """ Dedicated Exception class. """

    pass


def TLV_GET(attrs, attrNum, format):
    """ Get a tag-length-value encoded attribute. """
    attrView = attrs[attrNum]
    if format == 's':
        format = str(attrView.len) + format
    try:
        (result,) = struct.unpack_from(format, attrView.buf, attrView.offset)
    except TypeError:
        # Working around struct.unpack_from issue #10212
        (result,) = struct.unpack_from(format, str(bytearray(attrView.buf)), attrView.offset)
    return result


def TLV_PUT(attrs, attrNum, format, value):
    """ Put a tag-length-value encoded attribute. """
    attrView = attrs[attrNum]
    if format == 's':
        format = str(attrView.len) + format
    struct.pack_into(format, attrView.buf, attrView.offset, value)


def TLV_GET_BYTES(attrs, attrNum):
    """ Get a tag-length-value encoded attribute as bytes. """
    return TLV_GET(attrs, attrNum, 's')


def TLV_PUT_BYTES(attrs, attrNum, value):
    """ Put a tag-length-value encoded attribute as bytes. """
    TLV_PUT(attrs, attrNum, 's', value)


def TLV_GET_STRING(attrs, attrNum):
    """ Get a tag-length-value encoded attribute as a string. """
    return t.readString(TLV_GET_BYTES(attrs, attrNum))


def TLV_GET_UUID(attrs, attrNum):
    """ Get a tag-length-value encoded attribute as a UUID. """
    return btrfs.bytes2uuid(TLV_GET_BYTES(attrs, attrNum))


def TLV_GET_U64(attrs, attrNum):
    """ Get a tag-length-value encoded attribute as a U64. """
    return TLV_GET(attrs, attrNum, t.u64)


def replaceIDs(data, receivedUUID, receivedGen, parentUUID, parentGen):
    """ Parse and replace UUID and transid info in data stream. """
    if len(data) < 20:
        return data

    logger.debug(
        "Setting received %s/%d and parent %s/%d",
        receivedUUID, receivedGen or 0, parentUUID, parentGen or 0.
        )
    data = bytearray(data)  # Make data writable

    buf = ioctl.Buffer(data)
    header = buf.read(btrfs_stream_header)

    if header.magic != BTRFS_SEND_STREAM_MAGIC:
        raise ParseException("Didn't find '%s'" % (BTRFS_SEND_STREAM_MAGIC))

    logger.debug("Version: %d", header.version)

    if header.version > BTRFS_SEND_STREAM_VERSION:
        logger.warn("Unknown stream version: %d", header.version)

    cmdHeaderView = buf.peekView(btrfs_cmd_header.size)
    cmdHeader = buf.read(btrfs_cmd_header)

    logger.debug("Command: %d", cmdHeader.cmd)

    # Read the attributes

    attrs = {}
    attrDataView = buf.peekView(cmdHeader.len)
    attrData = buf.readBuffer(cmdHeader.len)

    while attrData.len > 0:
        attrHeader = attrData.read(btrfs_tlv_header)
        attrs[attrHeader.tlv_type] = attrData.readBuffer(attrHeader.tlv_len)

    def calcCRC():
        header = vars(cmdHeader)
        header['crc'] = 0

        # This works, but is slow
        # crc = crc32c.CRC_INIT ^ crc32c._MASK
        # crc = crc32c.crc_update(crc, btrfs_cmd_header.write(header))
        # crc = crc32c.crc_update(crc, attrDataView)
        # crc = crc32c.crc_finalize(crc)
        # crc = crc ^ crc32c._MASK

        # This works, and can be fast, when it used compiled extension
        crc = 0 ^ 0xffffffff
        crc = crc32c(btrfs_cmd_header.write(header), crc)
        crc = crc32c(attrDataView.tobytes(), crc)
        crc &= 0xffffffff
        crc = crc ^ 0xffffffff

        # This does *not* work
        # crc = 0 ^ 0xffffffff
        # crc = binascii.crc32(btrfs_cmd_header.write(header), crc)
        # crc = binascii.crc32(attrDataView, crc)
        # crc &= 0xffffffff
        # crc ^= 0xffffffff

        return crc

    crc = calcCRC()
    if cmdHeader.crc != crc:
        logger.warn(
            "Stored crc (%d) doesn't match calculated crc (%d)",
            cmdHeader.crc, crc,
            )

    # Dispatch based on cmd and attributes

    s = attrs

    def correct(attr, format, name, old, new, encode=None):
        if new is not None and new != old:
            logger.debug("Correcting %s from %s to %s", name, str(old), str(new))
            if encode is not None:
                new = encode(new)
            TLV_PUT(attrs, attr, format, new)

    def correctCRC():
        crc = calcCRC()
        if cmdHeader.crc != crc:
            logger.debug("Correcting CRC from %d to %d", cmdHeader.crc, crc)
            header = vars(cmdHeader)
            header['crc'] = crc
            cmdHeaderView[:] = btrfs_cmd_header.write(header).tostring()

    if cmdHeader.cmd == BTRFS_SEND_C_SUBVOL:
        path = TLV_GET_STRING(s, BTRFS_SEND_A_PATH, )
        uuid = TLV_GET_UUID(s, BTRFS_SEND_A_UUID, )
        ctransid = TLV_GET_U64(s, BTRFS_SEND_A_CTRANSID, )

        logger.debug('Subvol: %s/%d %s', uuid, ctransid, path)

        correct(
            BTRFS_SEND_A_UUID,
            's',
            'received UUID',
            uuid,
            receivedUUID,
            btrfs.uuid2bytes
        )
        correct(
            BTRFS_SEND_A_CTRANSID,
            t.u64,
            'received gen',
            ctransid,
            receivedGen
        )

    elif cmdHeader.cmd == BTRFS_SEND_C_SNAPSHOT:
        path = TLV_GET_STRING(s, BTRFS_SEND_A_PATH, )
        uuid = TLV_GET_UUID(s, BTRFS_SEND_A_UUID, )
        ctransid = TLV_GET_U64(s, BTRFS_SEND_A_CTRANSID, )
        clone_uuid = TLV_GET_UUID(s, BTRFS_SEND_A_CLONE_UUID, )
        clone_ctransid = TLV_GET_U64(s, BTRFS_SEND_A_CLONE_CTRANSID, )

        logger.debug(
            'Snapshot: %s/%d -> %s/%d %s',
            clone_uuid, clone_ctransid, uuid, ctransid, path
        )

        correct(
            BTRFS_SEND_A_UUID,
            's',
            'received UUID',
            uuid,
            receivedUUID,
            btrfs.uuid2bytes
        )
        correct(
            BTRFS_SEND_A_CTRANSID,
            t.u64,
            'received gen',
            ctransid,
            receivedGen
        )

        correct(
            BTRFS_SEND_A_CLONE_UUID,
            's',
            'parent UUID',
            clone_uuid,
            parentUUID,
            btrfs.uuid2bytes
        )
        correct(
            BTRFS_SEND_A_CLONE_CTRANSID,
            t.u64,
            'parent gen',
            clone_ctransid,
            parentGen
        )
    else:
        logger.warn("Didn't find volume UUID command")

    correctCRC()

    return data
